print("import: 'plotnine'")
import plotnine

print("import: 'plotnine.coords'")
import plotnine.coords

print("import: 'plotnine.data'")
import plotnine.data

print("import: 'plotnine.facets'")
import plotnine.facets

print("import: 'plotnine.geoms'")
import plotnine.geoms

print("import: 'plotnine.guides'")
import plotnine.guides

print("import: 'plotnine.mapping'")
import plotnine.mapping

print("import: 'plotnine.positions'")
import plotnine.positions

print("import: 'plotnine.scales'")
import plotnine.scales

print("import: 'plotnine.stats'")
import plotnine.stats

print("import: 'plotnine.themes'")
import plotnine.themes

